# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import scipp as sc

from .component import Component


class Chopper(Component):
    def __init__(
        self,
        frequency: sc.Variable,
        open: sc.Variable,
        close: sc.Variable,
        distance: sc.Variable,
        phase: sc.Variable,
        name: str = "",
    ):
        self.frequency = frequency
        self.open = open
        self.close = close
        self.distance = distance
        self.phase = phase
        self.name = name
        super().__init__()

    @property
    def omega(self) -> sc.Variable:
        return sc.constants.pi * (2.0 * sc.units.rad) * self.frequency

    @property
    def open_times(self) -> sc.Variable:
        return (
            self.open.to(unit='rad', copy=False) + self.phase.to(unit='rad', copy=False)
        ) / self.omega

    @property
    def close_times(self) -> sc.Variable:
        return (
            self.close.to(unit='rad', copy=False)
            + self.phase.to(unit='rad', copy=False)
        ) / self.omega

    def __repr__(self) -> str:
        return (
            f"Chopper(name={self.name}, distance={self.distance:c}, "
            f"frequency={self.frequency:c}, phase={self.phase:c}, "
            f"cutouts={len(self.open)})"
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import scipp as sc

from .component import Component


class Detector(Component):
    def __init__(self, distance: sc.Variable = 0.0, name: str = "detector"):
        self.distance = distance
        self.name = name
        super().__init__()

    def __repr__(self) -> str:
        return f"Detector(name={self.name}, distance={self.distance:c})"



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from itertools import chain
from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np
import scipp as sc
from matplotlib.collections import LineCollection

from .chopper import Chopper
from .detector import Detector
from .pulse import Pulse


class Model:
    def __init__(
        self,
        choppers: Union[Chopper, List[Chopper]],
        detectors: Union[Detector, List[Detector]],
        pulse: Pulse,
    ):
        self.choppers = choppers
        if not isinstance(self.choppers, (list, tuple)):
            self.choppers = [self.choppers]
        self.detectors = detectors
        if not isinstance(self.detectors, (list, tuple)):
            self.detectors = [self.detectors]
        self.pulse = pulse

    def run(self, npulses: int = 1):
        # TODO: ray-trace multiple pulses
        components = sorted(
            chain(self.choppers, self.detectors),
            key=lambda c: c.distance.value,
        )

        initial_mask = sc.ones(
            sizes=self.pulse.birth_times.sizes, unit=None, dtype=bool
        )
        for comp in components:
            comp._wavelengths = self.pulse.wavelengths
            t = self.pulse.birth_times + comp.distance / self.pulse.speeds
            comp._arrival_times = t
            if isinstance(comp, Detector):
                comp._mask = initial_mask
                continue
            m = sc.zeros(sizes=t.sizes, unit=None, dtype=bool)
            to = comp.open_times
            tc = comp.close_times
            for i in range(len(to)):
                m |= (t > to[i]) & (t < tc[i])
            combined = initial_mask & m
            comp._mask = combined
            comp._own_mask = ~m & initial_mask
            initial_mask = combined

    def _add_rays(self, ax, tofs, birth_times, distances, wavelengths=None):
        x0 = birth_times.to(unit='us', copy=False).values.reshape(-1, 1)
        x1 = tofs.to(unit='us', copy=False).values.reshape(-1, 1)
        y0 = np.zeros(x0.size).reshape(-1, 1)
        y1 = distances.values.reshape(-1, 1)
        segments = np.concatenate(
            (
                np.concatenate((x0, y0), axis=1).reshape(-1, 1, 2),
                np.concatenate((x1, y1), axis=1).reshape(-1, 1, 2),
            ),
            axis=1,
        )
        coll = LineCollection(segments)
        if wavelengths is not None:
            coll.set_cmap(plt.cm.gist_rainbow_r)
            coll.set_array(wavelengths.values)
            coll.set_norm(plt.Normalize(self.pulse.lmin.value, self.pulse.lmax.value))
            cbar = plt.colorbar(coll)
            cbar.ax.yaxis.set_label_coords(-0.9, 0.5)
            cbar.set_label('Wavelength (Å)')
        else:
            coll.set_color('lightgray')
        ax.add_collection(coll)

    def plot(self, max_rays: int = 1000, blocked_rays: int = 0, figsize=None) -> tuple:
        fig, ax = plt.subplots(figsize=figsize)
        furthest_detector = max(self.detectors, key=lambda d: d.distance)

        if blocked_rays > 0:
            inv_mask = ~furthest_detector._mask
            nrays = int(inv_mask.sum())
            if nrays > blocked_rays:
                inds = np.random.choice(nrays, size=blocked_rays, replace=False)
            else:
                inds = slice(None)
            birth_times = self.pulse.birth_times[inv_mask][inds]

            components = sorted(
                chain(self.choppers, [furthest_detector]),
                key=lambda c: c.distance.value,
            )
            dim = 'component'
            tofs = sc.concat(
                [comp._arrival_times[inv_mask][inds] for comp in components], dim=dim
            )
            distances = sc.concat(
                [
                    comp.distance.broadcast(sizes=birth_times.sizes)
                    for comp in components
                ],
                dim=dim,
            )
            masks = sc.concat(
                [sc.ones(sizes=birth_times.sizes, dtype=bool)]
                + [comp._mask[inv_mask][inds] for comp in components],
                dim=dim,
            )

            diff = sc.abs(masks[dim, 1:].to(dtype=int) - masks[dim, :-1].to(dtype=int))
            diff.unit = ''
            self._add_rays(
                ax=ax,
                tofs=(tofs * diff).max(dim=dim),
                birth_times=birth_times,
                distances=(distances * diff).max(dim=dim),
            )

        # Normal rays
        if max_rays > 0:
            tofs = furthest_detector.tofs.visible.data.coords['tof']
            if (max_rays is not None) and (len(tofs) > max_rays):
                inds = np.random.choice(len(tofs), size=max_rays, replace=False)
            else:
                inds = slice(None)
            birth_times = self.pulse.birth_times[furthest_detector._mask][inds]
            wavelengths = self.pulse.wavelengths[furthest_detector._mask][inds]
            distances = furthest_detector.distance.broadcast(sizes=birth_times.sizes)
            self._add_rays(
                ax=ax,
                tofs=tofs[inds],
                birth_times=birth_times,
                distances=distances,
                wavelengths=wavelengths,
            )

        tof_max = tofs.max().value
        # Plot choppers
        for ch in self.choppers:
            x0 = ch.open_times.to(unit='us').values
            x1 = ch.close_times.to(unit='us').values
            x = np.empty(3 * x0.size, dtype=x0.dtype)
            x[0::3] = x0
            x[1::3] = 0.5 * (x0 + x1)
            x[2::3] = x1
            x = np.concatenate([[0], x, [tof_max]])
            y = np.full_like(x, ch.distance.value)
            y[2::3] = None
            ax.plot(x, y, color="k")
            ax.text(
                tof_max, ch.distance.value, ch.name, ha="right", va="bottom", color="k"
            )

        # Plot detectors
        for det in self.detectors:
            ax.plot([0, tof_max], [det.distance.value] * 2, color="gray", lw=3)
            ax.text(
                0, det.distance.value, det.name, ha="left", va="bottom", color="gray"
            )

        # Plot pulse
        tmin = self.pulse.tmin.to(unit='us').value
        ax.plot(
            [tmin, self.pulse.tmax.to(unit='us').value],
            [0, 0],
            color="gray",
            lw=3,
        )
        ax.text(tmin, 0, "Pulse", ha="left", va="top", color="gray")

        ax.set_xlabel("Time-of-flight (us)")
        ax.set_ylabel("Distance (m)")
        fig.tight_layout()
        return fig, ax

    def __repr__(self) -> str:
        return (
            f"Model(choppers={self.choppers},\n      "
            f"detectors={self.detectors},\n      "
            f"pulse={self.pulse},\n      "
            f"neutrons={len(self.pulse.birth_times)})"
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Optional, Union

import plopp as pp
import scipp as sc


class Data:
    def __init__(self, data: sc.DataArray, dim: str):
        self._data = data
        self._dim = dim

    @property
    def data(self) -> sc.DataArray:
        return self._data

    def plot(self, bins: Union[int, sc.Variable] = 300):
        return self._data.hist({self._dim: bins}).plot()

    def __repr__(self) -> str:
        return f"Data(data={self._data})"


class ComponentData:
    def __init__(
        self,
        data: sc.Variable,
        mask: sc.Variable,
        dim: str,
        blocking: Optional[sc.Variable] = None,
    ):
        self._data = data
        self._mask = mask
        self._blocking = blocking
        self._dim = dim

    @property
    def visible(self) -> Data:
        a = self._data[self._mask]
        return Data(
            data=sc.DataArray(
                data=sc.ones(sizes=a.sizes, unit='counts'), coords={self._dim: a}
            ),
            dim=self._dim,
        )

    @property
    def blocked(self) -> Data:
        if self._blocking is None:
            return
        a = self._data[self._blocking]
        return Data(
            data=sc.DataArray(
                data=sc.ones(sizes=a.sizes, unit='counts'), coords={self._dim: a}
            ),
            dim=self._dim,
        )

    @property
    def data(self) -> sc.DataGroup:
        out = {'visible': self.visible.data}
        if self._blocking is not None:
            out['blocked'] = self.blocked.data
        return sc.DataGroup(out)

    def __repr__(self) -> str:
        return (
            f"ComponentData(visible={sc.sum(self._mask).value}, "
            f"blocked={sc.sum(~self._mask).value}, dim={self._dim})"
        )

    def plot(self, bins: Union[int, sc.Variable] = 300):
        if self._blocking is None:
            return self.visible.plot(bins=bins)
        visible = self.visible.data
        blocked = self.blocked.data
        if isinstance(bins, int):
            bins = sc.linspace(
                dim=self._dim,
                start=min(
                    visible.coords[self._dim].min(), blocked.coords[self._dim].min()
                ).value,
                stop=max(
                    visible.coords[self._dim].max(), blocked.coords[self._dim].max()
                ).value,
                num=bins,
                unit=visible.coords[self._dim].unit,
            )
        return pp.plot(
            {
                'visible': visible.hist({self._dim: bins}),
                'blocked': blocked.hist({self._dim: bins}),
            },
            color={'blocked': 'gray'},
        )


class Component:
    def __init__(self):
        self._arrival_times = None
        self._wavelengths = None
        self._mask = None
        self._own_mask = None

    @property
    def tofs(self) -> sc.Variable:
        return ComponentData(
            data=self._arrival_times.to(unit='us'),
            mask=self._mask,
            blocking=self._own_mask,
            dim='tof',
        )

    @property
    def wavelengths(self) -> sc.Variable:
        return ComponentData(
            data=self._wavelengths,
            mask=self._mask,
            blocking=self._own_mask,
            dim='wavelength',
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

# flake8: noqa F401

from .chopper import Chopper
from .detector import Detector
from .model import Model
from .pulse import Pulse



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import scipp as sc

from . import facilities
from .utils import wavelength_to_speed


class Pulse:
    def __init__(
        self,
        tmin: Optional[sc.Variable] = None,
        tmax: Optional[sc.Variable] = None,
        lmin: Optional[sc.Variable] = None,
        lmax: Optional[sc.Variable] = None,
        neutrons: int = 1_000_000,
        kind: Optional[str] = None,
        p_wav: Optional[np.ndarray] = None,
        p_time: Optional[np.ndarray] = None,
        sampling_resolution: int = 10000,
    ):
        self.kind = kind
        self.neutrons = neutrons
        self.tmin = tmin
        self.tmax = tmax
        self.lmin = lmin
        self.lmax = lmax
        self.p_wav = p_wav
        self.p_time = p_time
        self.sampling_resolution = sampling_resolution
        self.generate()

    def generate(self, neutrons: Optional[int] = None):
        if neutrons is not None:
            self.neutrons = neutrons
        if self.kind is not None:
            params = getattr(facilities, self.kind)
            if self.tmin is None:
                self.tmin = params['time'].coords['time'].min()
            if self.tmax is None:
                self.tmax = params['time'].coords['time'].max()
            if self.lmin is None:
                self.lmin = params['wavelength'].coords['wavelength'].min()
            if self.lmax is None:
                self.lmax = params['wavelength'].coords['wavelength'].max()
            self.tmin = self.tmin.to(unit='s')
            self.tmax = self.tmax.to(unit='s')
            self.lmin = self.lmin.to(unit='angstrom')
            self.lmax = self.lmax.to(unit='angstrom')

            x_time = np.linspace(
                self.tmin.value, self.tmax.value, self.sampling_resolution
            )
            x_wav = np.linspace(
                self.lmin.value, self.lmax.value, self.sampling_resolution
            )
            p_time = np.interp(
                x_time,
                params['time'].coords['time'].to(unit='s').values,
                params['time'].values,
            )
            p_time /= p_time.sum()
            p_wav = np.interp(
                x_wav,
                params['wavelength'].coords['wavelength'].to(unit='angstrom').values,
                params['wavelength'].values,
            )
            p_wav /= p_wav.sum()
        else:
            self.tmin = self.tmin.to(unit='s')
            self.tmax = self.tmax.to(unit='s')
            self.lmin = self.lmin.to(unit='angstrom')
            self.lmax = self.lmax.to(unit='angstrom')
            p_time = self.p_time
            p_wav = self.p_wav

        if p_time is not None:
            if self.kind is None:
                x_time = np.linspace(self.tmin.value, self.tmax.value, len(p_time))
            self.birth_times = np.random.choice(x_time, size=self.neutrons, p=p_time)
        else:
            self.birth_times = np.random.uniform(
                self.tmin.value, self.tmax.value, self.neutrons
            )
        if p_wav is not None:
            if self.kind is None:
                x_wav = np.linspace(self.lmin.value, self.lmax.value, len(p_wav))
            self.wavelengths = np.random.choice(x_wav, size=self.neutrons, p=p_wav)
        else:
            self.wavelengths = np.random.uniform(self.lmin, self.lmax, self.neutrons)

        self.birth_times = sc.array(dims=['event'], values=self.birth_times, unit='s')
        self.wavelengths = sc.array(
            dims=['event'], values=self.wavelengths, unit='angstrom'
        )
        self.speeds = wavelength_to_speed(self.wavelengths)

    @property
    def duration(self) -> float:
        return self.tmax - self.tmin

    def plot(self, bins: int = 300) -> tuple:
        fig, ax = plt.subplots(1, 2)
        self.birth_times.hist(time=bins).plot(ax=ax[0])
        self.wavelengths.hist(wavelength=bins).plot(ax=ax[1])
        size = fig.get_size_inches()
        fig.set_size_inches(size[0] * 2, size[1])
        fig.tight_layout()
        return fig, ax

    def __repr__(self) -> str:
        return (
            f"Pulse(tmin={self.tmin:c}, tmax={self.tmax:c}, lmin={self.lmin:c}, "
            f"lmax={self.lmax:c}, neutrons={self.neutrons}, kind={self.kind})"
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import scipp as sc
import scipp.constants as const

alpha = const.m_n / const.h


def speed_to_wavelength(x: sc.Variable, unit: str = 'angstrom') -> sc.Variable:
    return (1.0 / (alpha * x)).to(unit=unit)


def wavelength_to_speed(x: sc.Variable, unit: str = 'm/s') -> sc.Variable:
    return (1.0 / (alpha * x)).to(unit=unit)


def speed_to_energy(x: sc.Variable, unit='meV') -> sc.Variable:
    return (const.m_n * x * x).to(unit=unit)


def energy_to_speed(x: sc.Variable, unit='m/s') -> sc.Variable:
    return sc.sqrt(x / const.m_n).to(unit=unit)
